@@ -1,27 +1,33 @@
+import os
+import re
 import sys
 import time
 import json
 import signal
+import platform
 import logging
 import datetime
 import subprocess
 import pandas as pd
-from os.path import join, split
-from d3m_interface.basic_ta3 import BasicTA3
-from d3m_interface.visualization import plot_metadata, plot_comparison_pipelines, plot_text_summary, plot_text_explanation
-from d3m_interface.data_converter import is_d3m_format, dataset_to_d3m, d3mtext_to_dataframe, copy_folder, to_d3m_json
-from d3m_interface.pipeline import Pipeline
+import d3m_interface.visualization as vis
+from d3m_interface.data_converter import is_d3m_format, is_d3m_collection, dataset_to_d3m, d3mtext_to_dataframe, to_d3m_json
 from d3m_interface.confidence_calculator import create_confidence_pipeline
-from d3m.metadata.problem import PerformanceMetric
+from d3m_interface.utils import copy_folder, fix_path_for_docker
+from d3m_interface.grpc_client import GrpcClient
+from d3m_interface.pipeline import Pipeline
 from threading import Thread
+from os.path import join, split, exists
+from posixpath import join as pjoin
 from IPython.core.getipython import get_ipython
+from d3m.metadata.problem import PerformanceMetric
+from d3m.utils import compute_digest
 
 
 logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s', stream=sys.stdout)
 logger = logging.getLogger(__name__)
 
 
-TA2_DOCKER_IMAGES = {'AlphaD3M': 'registry.gitlab.com/vida-nyu/d3m/alphad3m:latest',
+TA2_DOCKER_IMAGES = {'AlphaD3M': 'registry.gitlab.com/vida-nyu/d3m/alphad3m:master',
                      'CMU': 'registry.gitlab.com/sray/cmu-ta2:latest',
                      'SRI': 'registry.gitlab.com/daraghhartnett/autoflow:latest',
                      'TAMU': 'dmartinez05/tamuta2:latest'}
@@ -33,34 +39,145 @@ IGNORE_SUMMARY_PRIMITIVES = {'d3m.primitives.data_transformation.construct_predi
                              'd3m.primitives.data_transformation.column_parser.Common'}
 
 
-def kernel_interrupt_handler(signal, frame):
-    logger.info('KeyboardInterrupt signal received, ending session...')
-    subprocess.call(['docker', 'stop', 'ta2_container'])
-    logger.info('Session ended!')
+class DockerRuntime:
+    dataset_in_container = '/input/dataset'
+    output_in_container = '/output'
 
-    raise KeyboardInterrupt
+    def __init__(self, image, dataset, output_folder, port=45042):
+        process_returncode = 0
+        while process_returncode == 0:
+            # Force to stop the docker container
+            process_returncode = subprocess.call(['docker', 'stop', 'ta2_container'])
+            time.sleep(2)
+
+        logger.info("Creating Docker container...")
+        self.proc = subprocess.Popen(
+            [
+                'docker', 'run', '--rm',
+                '--name', 'ta2_container',
+                '-p', '%d:45042' % port,
+                '-e', 'D3MRUN=ta2ta3',
+                '-e', 'D3MINPUTDIR=/input',
+                '-e', 'D3MOUTPUTDIR=/output',
+                '-e', 'D3MSTATICDIR=/output',  # TODO: Temporal assignment for D3MSTATICDIR env variable
+                '-v', '%s:/input/dataset/' % fix_path_for_docker(dataset),
+                '-v', '%s:/output' % fix_path_for_docker(output_folder),
+                image,
+            ],
+        )
+
+    def run_command(self, args):
+        cmd = ['docker', 'exec', 'ta2_container']
+        cmd.extend(args)
+        process = subprocess.Popen(
+            cmd,
+            stderr=subprocess.PIPE,
+        )
+        _, stderr = process.communicate()
+
+        if process.returncode != 0:
+            raise RuntimeError(stderr.decode())
+
+    def close(self):
+        subprocess.call(['docker', 'stop', 'ta2_container'])
+
+
+class SingularityRuntime:
+    dataset_in_container = '/input/dataset'
+    output_in_container = '/output'
+
+    def __init__(self, image, dataset, output_folder, port=45042):
+        if port != 45042:
+            raise ValueError(
+                "There is currently no way to change the port used by the "
+                + "AutoML system when using Singularity"
+            )
+
+        process_returncode = 0
+        while process_returncode == 0:
+            # Force to stop the docker container
+            process_returncode = subprocess.call(['singularity', 'instance', 'stop', 'ta2_container'])
+            time.sleep(2)
+
+        os.makedirs(output_folder, exist_ok=True)
+        logger.info("Creating Singularity instance...")
+        process = subprocess.Popen(
+            [
+                'singularity', 'instance', 'start',
+                '--writable-tmpfs',
+                '--env', 'D3MRUN=ta2ta3',
+                '--env', 'D3MINPUTDIR=/input',
+                '--env', 'D3MOUTPUTDIR=/output',
+                '--env', 'D3MSTATICDIR=/output',  # TODO: Temporal assignment for D3MSTATICDIR env variable
+                '--bind', '%s:/input/dataset/' % dataset,
+                '--bind', '%s:/output' % output_folder,
+                'docker://' + image,
+                'ta2_container',
+            ],
+            stderr=subprocess.PIPE,
+        )
+        _, stderr = process.communicate()
+
+        if process.returncode != 0:
+            raise RuntimeError(stderr.decode())
+
+        logger.info("Instance created, running command...")
+        self.proc = subprocess.Popen(
+            [
+                'singularity', 'run',
+                '--writable-tmpfs',
+                '--env', 'D3MRUN=ta2ta3',
+                '--env', 'D3MINPUTDIR=/input',
+                '--env', 'D3MOUTPUTDIR=/output',
+                '--env', 'D3MSTATICDIR=/output',  # TODO: Temporal assignment for D3MSTATICDIR env variable
+                '--bind', '%s:/input/dataset/' % dataset,
+                '--bind', '%s:/output' % output_folder,
+                'instance://ta2_container',
+            ],
+        )
+
+    def run_command(self, args):
+        cmd = ['singularity', 'exec', 'instance://ta2_container']
+        cmd.extend(args)
+        process = subprocess.Popen(
+            cmd,
+            stderr=subprocess.PIPE,
+        )
+        _, stderr = process.communicate()
+
+        if process.returncode != 0:
+            raise RuntimeError(stderr.decode())
+
+    def close(self):
+        subprocess.call(['singularity', 'instance', 'stop', 'ta2_container'])
 
 
 class AutoML:
-
-    def __init__(self, output_folder, ta2_id='AlphaD3M'):
+    def __init__(self, output_folder, ta2_id='AlphaD3M', container_runtime='docker', grpc_port=45042):
         """Create/instantiate an AutoML object
 
         :param output_folder: Path to the output directory
         :param ta2_id: AutoML system name. It makes reference to the AutoML docker image. The provided AutoML systems
             are the following: `AlphaD3M, CMU, SRI, TAMU`
+        :param container_runtime: The container runtime to use, either 'docker' or 'singularity'
         """
         if ta2_id not in TA2_DOCKER_IMAGES:
             raise ValueError('Unknown "%s" AutoML, you should choose among: [%s]' % (ta2_id, ', '.join(TA2_DOCKER_IMAGES)))
 
         self.output_folder = output_folder
+        if container_runtime == 'docker':
+            self.container_runtime = DockerRuntime
+        elif container_runtime == 'singularity':
+            self.container_runtime = SingularityRuntime
+        else:
+            raise ValueError("Unknown container runtime %r" % container_runtime)
         self.ta2_id = ta2_id
         self.pipelines = {}
         self.ta2 = None
         self.ta3 = None
         self.dataset = None
-        self.leaderboard = None
         self.problem_config = None
+        self.port = grpc_port
 
     def search_pipelines(self, dataset, time_bound, time_bound_run=5, target=None, metric=None, task_keywords=None,
                          method='holdout', stratified=True, shuffle=True, folds=10, train_ratio=0.70, random_seed=0,
@@ -88,11 +205,8 @@ class AutoML:
         :param train_ratio: Represent the proportion of the dataset to include in the train split
         :param random_seed: The number seed used by the random generator
         :param kwargs: Different arguments for problem's settings (e.g. pos_label for binary problems using F1)
-        :returns: List of pipelines
         """
         suffix = 'TRAIN'
-        dataset_in_container = '/input/dataset/'
-        signal.signal(signal.SIGINT, kernel_interrupt_handler)
 
         if not is_d3m_format(dataset, suffix):
             self.problem_config = {'target_column': target, 'metric': metric, 'task_keywords': task_keywords,
@@ -102,24 +216,29 @@ class AutoML:
         self.dataset = split(dataset)[0]
         self.start_ta2()
         search_id = None
-        signal.signal(signal.SIGALRM, lambda signum, frame: self.ta3.do_stop_search(search_id))
-        signal.alarm(time_bound * 60)
-        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        if platform.system() != 'Windows':
+            signal.signal(signal.SIGALRM, lambda signum, frame: self.ta3.stop_search(search_id))
+            signal.alarm(time_bound * 60)
+        train_dataset_d3m = pjoin(self.ta2.dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
         problem_path = join(dataset, 'problem_TRAIN/problemDoc.json')
         start_time = datetime.datetime.utcnow()
-        pipelines = self.ta3.do_search(train_dataset_d3m, problem_path, time_bound, time_bound_run)
+        try:
+            search_id = self.ta3.search_solutions(train_dataset_d3m, problem_path, time_bound, time_bound_run)
+            pipelines = self.ta3.get_solutions(search_id)
+        except KeyboardInterrupt:
+            self.ta3.stop_search(search_id)
+            raise
 
         jobs = []
-
         for pipeline in pipelines:
             end_time = datetime.datetime.utcnow()
             try:
-                pipeline_json = self.ta3.do_describe(pipeline['id'])
-            except:
-                logger.warning('Pipeline id=%s could not be decoded' % pipeline['id'])
+                pipeline_json = self.ta3.describe_solution(pipeline['id'])
+            except Exception as e:
+                logger.warning('Pipeline id=%s could not be decoded' % pipeline['id'], exc_info=e)
                 continue
             summary_pipeline = self.get_summary_pipeline(pipeline_json)
-            search_id = pipeline['search_id']
+            pipeline['search_id'] = search_id
             pipeline['json_representation'] = pipeline_json
             pipeline['summary'] = summary_pipeline
             duration = str(end_time - start_time)
@@ -140,18 +259,8 @@ class AutoML:
         else:
             logger.info('Search completed, no pipelines found!')
 
-        if len(self.pipelines) > 0:
-            leaderboard = []
-            sorted_pipelines = sorted(self.pipelines.values(), key=lambda x: x['normalized_score'], reverse=True)
-            metric = sorted_pipelines[0]['metric']
-            for position, pipeline_data in enumerate(sorted_pipelines, 1):
-                leaderboard.append([position, pipeline_data['id'], pipeline_data['summary'],  pipeline_data['score']])
-
-            self.leaderboard = pd.DataFrame(leaderboard, columns=['ranking', 'id', 'summary', metric])
-
-        signal.alarm(0)
-
-        return self.pipelines.values()
+        if platform.system() != 'Windows':
+            signal.alarm(0)
 
     def train(self, pipeline_id, expose_outputs=None):
         """Train a model using an specific ML pipeline
@@ -160,9 +269,9 @@ class AutoML:
         :param expose_outputs: The output of the pipeline steps. If None, it doesn't expose any output of the steps.
             If str, should be 'all' to shows the output of each step in the pipeline, If list, it should contain the
             ids of the steps, e.g. 'steps.2.produce'
-        :returns: A dictionary that contains the id and fitted model with/without the pipeline step outputs
+        :returns: An id of the fitted pipeline with/without the pipeline step outputs
         """
-        dataset_in_container = '/input/dataset/'
+        # TODO: It should receive the path to the train dataset as an optional parameter
 
         if pipeline_id not in self.pipelines:
             raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
@@ -176,17 +285,8 @@ class AutoML:
                 for id_output in step['outputs']:
                     expose_outputs.append('steps.%d.%s' % (index, id_output['id']))
 
-        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
-        fitted_pipeline_id, pipeline_step_outputs = self.ta3.do_train(pipeline_id, train_dataset_d3m, expose_outputs)
-        fitted_pipeline = None
-        # TODO: Implement other method to save the fitted pipeline
-        # try:
-        #     fitted_solution_uri = self.ta3.do_save_fitted_solution(fitted_pipeline_id)
-        #     fitted_solution_uri = fitted_solution_uri.replace('file:///output/', '')
-        #     with open(join(self.output_folder, fitted_solution_uri), 'rb') as fin:
-        #         fitted_pipeline = pickle.load(fin)
-        # except Exception as e:
-        #     logger.warning('Fitted pipeline id=%s could not be loaded' % pipeline_id, exc_info=e)
+        train_dataset_d3m = pjoin(self.ta2.dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        fitted_pipeline_id, pipeline_step_outputs = self.ta3.train_solution(pipeline_id, train_dataset_d3m, expose_outputs)
 
         for step_id, step_csv_uri in pipeline_step_outputs.items():
             if not step_csv_uri.startswith('file://'):
@@ -197,18 +297,17 @@ class AutoML:
             pipeline_step_outputs[step_id] = step_dataframe
 
         self.pipelines[pipeline_id]['fitted_id'] = fitted_pipeline_id
-        model = {'id': pipeline_id, 'fitted': fitted_pipeline}
         logger.info('Training finished!')
 
         if len(expose_outputs) == 0:
-            return model
+            return fitted_pipeline_id
 
-        return model, pipeline_step_outputs
+        return fitted_pipeline_id, pipeline_step_outputs
 
-    def test(self, model, test_dataset, expose_outputs=None, calculate_confidence=False):
+    def test(self, pipeline_id, test_dataset, expose_outputs=None, calculate_confidence=False):
         """Test a model
 
-        :param model: Dict that contains the id and fitted model
+        :param pipeline_id: The id of a fitted pipeline
         :param test_dataset: Path to dataset. It supports D3M dataset, and CSV file
         :param expose_outputs: The output of the pipeline steps. If None, it doesn't expose any output of the steps.
             If str, should be 'all' to shows the output of each step in the pipeline, If list, it should contain the
@@ -217,21 +316,20 @@ class AutoML:
         :returns: A dataframe that contains the predictions with/without the pipeline step outputs
         """
         suffix = 'TEST'
-        dataset_in_container = '/input/dataset/'
-        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
-        problem_path = join(dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
+        dataset_in_container = self.ta2.dataset_in_container
+        train_dataset_d3m = pjoin(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        problem_path = pjoin(dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
 
         if not is_d3m_format(test_dataset, suffix):
             dataset_to_d3m(test_dataset, self.output_folder, self.problem_config, suffix)
         elif test_dataset != join(self.dataset, 'TEST'):  # Special case for D3M test dataset with different path
             destination_path = join(self.output_folder, 'temp', 'dataset_d3mformat', 'TEST')
             if test_dataset != destination_path:
-                copy_folder(test_dataset, destination_path)
-            dataset_in_container = '/output/temp/dataset_d3mformat/'
+                copy_folder(test_dataset, destination_path, True)
+            dataset_in_container = pjoin(self.ta2.output_in_container,  'temp', 'dataset_d3mformat')
 
         logger.info('Testing model...')
-        pipeline_id = model['id']
-        test_dataset_d3m = join(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
+        test_dataset_d3m = pjoin(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
 
         if calculate_confidence:
             # The only way to get the confidence is through the CLI utility, TA3TA2 API doesn't support it
@@ -241,29 +339,21 @@ class AutoML:
             with open(join(self.output_folder, '%s.json' % pipeline_id), 'w') as fout:
                 json.dump(confidence_pipeline, fout)  # Save temporally the json pipeline
 
-            pipeline_path = join('/output/', '%s.json' % pipeline_id)
-            output_csv_path = join('/output/', 'fit_produce_%s.csv' % pipeline_id)
+            pipeline_path = pjoin(self.ta2.output_in_container, '%s.json' % pipeline_id)
+            output_csv_path = pjoin(self.ta2.output_in_container, 'fit_produce_%s.csv' % pipeline_id)
 
-            process = subprocess.Popen(
-                [
-                    'docker', 'exec', 'ta2_container',
-                    'python3', '-m', 'd3m',
-                    'runtime',
-                    '--context', 'TESTING',
-                    '--random-seed', '0',
-                    'fit-produce',
-                    '--pipeline', pipeline_path,
-                    '--problem', problem_path,
-                    '--input', train_dataset_d3m,
-                    '--test-input', test_dataset_d3m,
-                    '--output', output_csv_path,
-                ],
-                stderr=subprocess.PIPE
-            )
-            _, stderr = process.communicate()
-
-            if process.returncode != 0:
-                raise RuntimeError(stderr.decode())
+            self.ta2.run_command([
+                'python3', '-m', 'd3m',
+                'runtime',
+                '--context', 'TESTING',
+                '--random-seed', '0',
+                'fit-produce',
+                '--pipeline', pipeline_path,
+                '--problem', problem_path,
+                '--input', train_dataset_d3m,
+                '--test-input', test_dataset_d3m,
+                '--output', output_csv_path,
+            ])
 
             result_path = join(self.output_folder, 'fit_produce_%s.csv' % pipeline_id)
             predictions = pd.read_csv(result_path)
@@ -285,7 +375,7 @@ class AutoML:
         if 'outputs.0' not in expose_outputs:
             expose_outputs.append('outputs.0')
 
-        pipeline_step_outputs = self.ta3.do_test(fitted_pipeline_id, test_dataset_d3m, expose_outputs)
+        pipeline_step_outputs = self.ta3.test_solution(fitted_pipeline_id, test_dataset_d3m, expose_outputs)
 
         for step_id, step_csv_uri in pipeline_step_outputs.items():
             if not step_csv_uri.startswith('file://'):
@@ -306,12 +396,11 @@ class AutoML:
     def score(self, pipeline_id, test_dataset):
         """Compute a proper score of the model
 
-        :param pipeline_id: A id of a pipeline or a Pipeline object
+        :param pipeline_id: The id of a pipeline or a Pipeline object
         :param test_dataset: Path to dataset. It supports D3M dataset, and CSV file
         :returns: A tuple holding metric name and score value
         """
         suffix = 'SCORE'
-        dataset_in_container = '/input/dataset/'
 
         if not is_d3m_format(test_dataset, suffix):
             # D3M format needs TEST and SCORE directories
@@ -325,39 +414,33 @@ class AutoML:
             if pipeline_id not in self.pipelines:
                 raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
             pipeline_json = self.pipelines[pipeline_id]['json_representation']
+        else:
+            raise TypeError("pipeline_id should be a Pipeline or str object")
 
         with open(join(self.output_folder, '%s.json' % pipeline_id), 'w') as fout:
             json.dump(pipeline_json, fout)  # Save temporally the json pipeline
 
-        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
-        test_dataset_d3m = join(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
-        score_dataset_d3m = join(dataset_in_container, 'SCORE/dataset_SCORE/datasetDoc.json')
-        problem_path = join(dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
-        pipeline_path = join('/output/', '%s.json' % pipeline_id)
-        output_csv_path = join('/output/', 'fit_score_%s.csv' % pipeline_id)
+        train_dataset_d3m = pjoin(self.ta2.dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        test_dataset_d3m = pjoin(self.ta2.dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
+        score_dataset_d3m = pjoin(self.ta2.dataset_in_container, 'SCORE/dataset_SCORE/datasetDoc.json')
+        problem_path = pjoin(self.ta2.dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
+        pipeline_path = pjoin(self.ta2.output_in_container, '%s.json' % pipeline_id)
+        output_csv_path = pjoin(self.ta2.output_in_container, 'fit_score_%s.csv' % pipeline_id)
 
-        #  TODO: Use TA2TA3 API to score
-        process = subprocess.Popen(
-            [
-                'docker', 'exec', 'ta2_container',
-                'python3', '-m', 'd3m',
-                'runtime',
-                '--context', 'TESTING',
-                '--random-seed', '0',
-                'fit-score',
-                '--pipeline', pipeline_path,
-                '--problem', problem_path,
-                '--input', train_dataset_d3m,
-                '--test-input', test_dataset_d3m,
-                '--score-input', score_dataset_d3m,
-                '--scores', output_csv_path
-            ],
-            stderr=subprocess.PIPE
-        )
-        _, stderr = process.communicate()
-
-        if process.returncode != 0:
-            raise RuntimeError(stderr.decode())
+        # TODO: Use TA2TA3 API to score
+        self.ta2.run_command([
+            'python3', '-m', 'd3m',
+            'runtime',
+            '--context', 'TESTING',
+            '--random-seed', '0',
+            'fit-score',
+            '--pipeline', pipeline_path,
+            '--problem', problem_path,
+            '--input', train_dataset_d3m,
+            '--test-input', test_dataset_d3m,
+            '--score-input', score_dataset_d3m,
+            '--scores', output_csv_path,
+        ])
 
         result_path = join(self.output_folder, 'fit_score_%s.csv' % pipeline_id)
         df = pd.read_csv(result_path)
@@ -366,6 +449,60 @@ class AutoML:
 
         return metric, score
 
+    def save_pipeline(self, pipeline_id, output_folder):
+        """Save a pipeline on disk
+
+        :param pipeline_id: The id of the pipeline to be saved
+        :param output_folder: Path to the folder where the pipeline will be saved
+        """
+        if pipeline_id not in self.pipelines:
+            raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
+
+        solution_uri = self.ta3.save_solution(pipeline_id)
+        folder_src_path = join(self.output_folder, solution_uri.replace('file:///output/', ''))
+        folder_dst_path = join(output_folder, pipeline_id)
+        copy_folder(folder_src_path, folder_dst_path, True)
+        pipeline_run = {
+            'metric': self.pipelines[pipeline_id]['metric'],
+            'score': self.pipelines[pipeline_id]['score'],
+            'normalized_score': self.pipelines[pipeline_id]['normalized_score'],
+            'start_time': self.pipelines[pipeline_id]['start_time'],
+            'end_time': self.pipelines[pipeline_id]['end_time'],
+            'search_id': self.pipelines[pipeline_id]['search_id'],
+            'summary': self.pipelines[pipeline_id]['summary'],
+            'json_representation': self.pipelines[pipeline_id]['json_representation'],
+            'dataset_path': self.dataset
+        }
+
+        with open(join(folder_dst_path, 'run.json'), 'w') as fout:
+            json.dump(pipeline_run, fout, indent=2)
+
+        logger.info('Pipeline saved at "%s"' % folder_dst_path)
+
+    def load_pipeline(self, pipeline_path):
+        """Load a previous saved pipeline
+
+        :param pipeline_path: Path to the folder where the pipeline is saved
+        """
+        pipeline_run_path = join(pipeline_path, 'run.json')
+        if not exists(pipeline_run_path):
+            raise ValueError('Can not load pipelines from "%s". File run.json does not exist' % pipeline_path)
+
+        with open(pipeline_run_path) as fin:
+            pipeline_run = json.load(fin)
+
+        if self.ta2 is None:
+            self.dataset = pipeline_run['dataset_path']
+            self.start_ta2()
+
+        folder_dst_path = join(self.output_folder, 'temp', 'loaded_pipelines')
+        copy_folder(pipeline_path, folder_dst_path, True)
+        pipeline_path_container = pjoin(self.ta2.output_in_container, 'temp', 'loaded_pipelines')
+        id_pipeline = self.ta3.load_solution(pipeline_path_container)
+        pipeline_run['id'] = id_pipeline
+        self.pipelines[id_pipeline] = pipeline_run
+        logger.info('Pipeline id=%s loaded!' % id_pipeline)
+
     def create_pipelineprofiler_inputs(self, test_dataset=None, source_name=None):
         """Create an proper input supported by PipelineProfiler based on the pipelines generated by a TA2 system
 
@@ -387,7 +524,7 @@ class AutoML:
             if pipeline['id'] not in pipeline_ids:
                 pipeline_ids.add(pipeline['id'])
                 if 'digest' not in pipeline['json_representation']:
-                    pipeline['json_representation']['digest'] = pipeline['id']  # TODO: Compute digest
+                    pipeline['json_representation']['digest'] = compute_digest(pipeline['json_representation'])
 
                 pipeline_score = [{'metric': {'metric': pipeline['metric']}, 'value': pipeline['score'],
                                    'normalized': pipeline['normalized_score']}]
@@ -400,8 +537,8 @@ class AutoML:
                     start_time = datetime.datetime.utcnow().isoformat() + 'Z'
                     try:
                         metric, score,  = self.score(pipeline['id'], test_dataset)
-                    except:
-                        logger.warning('Pipeline id=%s could not be scored' % pipeline['id'])
+                    except Exception:
+                        logger.warning('Pipeline id=%s could not be scored' % pipeline['id'], exc_info=True)
                         continue
                     end_time = datetime.datetime.utcnow().isoformat() + 'Z'
                     normalized_score = PerformanceMetric[metric.upper()].normalize(score)
@@ -428,6 +565,24 @@ class AutoML:
 
         return profiler_inputs
 
+    def create_textanalizer_inputs(self, dataset, text_column, label_column, positive_label=1, negative_label=0):
+        """Create an proper input supported by VisualTextAnalyzer
+
+        :param dataset: Path to dataset.  It supports D3M dataset, and CSV file
+        :param text_column: Name of the column that contains the texts
+        :param label_column: Name of the column that contains the classes
+        :param positive_label: Label for the positive class
+        :param negative_label: Label for the negative class
+        """
+        suffix = split(dataset)[-1]
+
+        if is_d3m_format(dataset, suffix):
+            dataframe = d3mtext_to_dataframe(dataset, text_column)
+        else:
+            dataframe = pd.read_csv(dataset, index_col=False)
+
+        return vis.get_words_entities(dataframe, text_column, label_column, positive_label, negative_label)
+
     def export_pipeline_code(self, pipeline_id, ipython_cell=True):
         """Converts a Pipeline Description to an executable Python script
 
@@ -500,36 +655,42 @@ class AutoML:
         """
         logger.info('Ending session...')
         if self.ta2 is not None:
-            subprocess.call(['docker', 'stop', 'ta2_container'])
+            self.ta2.close()
+            self.ta2 = None
 
         logger.info('Session ended!')
 
     def start_ta2(self):
-        logger.info('Initializing %s AutoML...', self.ta2_id)
-        subprocess.call(['docker', 'stop', 'ta2_container'])
+        if self.ta2 is not None:
+            self.end_session()
 
-        self.ta2 = subprocess.Popen(
-            [
-                'docker', 'run', '--rm',
-                '--name', 'ta2_container',
-                '-p', '45042:45042',
-                '-e', 'D3MRUN=ta2ta3',
-                '-e', 'D3MINPUTDIR=/input',
-                '-e', 'D3MOUTPUTDIR=/output',
-                '-e', 'D3MSTATICDIR=/output',  # TODO: Temporal assignment for D3MSTATICDIR env variable
-                '-v', '%s:/input/dataset/' % self.dataset,
-                '-v', '%s:/output' % self.output_folder,
-                TA2_DOCKER_IMAGES[self.ta2_id]
-            ]
+        logger.info('Initializing %s AutoML...', self.ta2_id)
+
+        self.ta2 = self.container_runtime(
+            TA2_DOCKER_IMAGES[self.ta2_id],
+            self.dataset,
+            self.output_folder,
+            self.port,
         )
+
         time.sleep(4)  # Wait for TA2
+
+        host = 'localhost'
+        if os.environ.get('DOCKER_HOST'):
+            m = re.match('tcp://([^:/]*):[0-9]+', os.environ['DOCKER_HOST'])
+            if m is not None:
+                host = m.group(1)
+            else:
+                logger.warning("Can't understand DOCKER_HOST")
+
+        logger.info("Connecting via gRPC to %s:%d...", host, self.port)
         while True:
             try:
-                self.ta3 = BasicTA3()
+                self.ta3 = GrpcClient(host, self.port)
                 self.ta3.do_hello()
                 logger.info('%s AutoML initialized!', self.ta2_id)
                 break
-            except:
+            except Exception:
                 if self.ta3.channel is not None:
                     self.ta3.channel.close()
                     self.ta3 = None
@@ -540,8 +701,8 @@ class AutoML:
                         train_ratio, random_seed):
         try:
             pipeline['start_time'] = datetime.datetime.utcnow().isoformat() + 'Z'
-            score_data = self.ta3.do_score(pipeline['id'], train_dataset, problem_path, method, stratified, shuffle,
-                                           folds, train_ratio, random_seed)
+            score_data = self.ta3.score_solutions(pipeline['id'], train_dataset, problem_path, method, stratified, shuffle,
+                                                  folds, train_ratio, random_seed)
             logger.info('Scored pipeline id=%s, %s=%s' % (pipeline['id'], score_data['metric'], score_data['score']))
             pipeline['end_time'] = datetime.datetime.utcnow().isoformat() + 'Z'
             pipeline['score'] = score_data['score']
@@ -565,31 +726,50 @@ class AutoML:
     def plot_leaderboard(self):
         """Plot pipelines' leaderboard
         """
-        return self.leaderboard.style.hide_index()
+        leaderboard_data = []
+        metric = 'metric'
 
-    def plot_summary_dataset(self, dataset):
+        if len(self.pipelines) > 0:
+            sorted_pipelines = sorted(self.pipelines.values(), key=lambda x: x['normalized_score'], reverse=True)
+            metric = sorted_pipelines[0]['metric']
+            for position, pipeline in enumerate(sorted_pipelines, 1):
+                leaderboard_data.append([position, pipeline['id'], pipeline['summary'],  pipeline['score']])
+
+        leaderboard = pd.DataFrame(leaderboard_data, columns=['ranking', 'id', 'summary', metric])
+
+        return leaderboard.style.hide_index()
+
+    def plot_summary_dataset(self, dataset, text_column=None):
         """Plot histograms of the dataset
 
         :param dataset: Path to dataset.  It supports D3M dataset, and CSV file
+        :param text_column: Name of the column that contains the texts. Only needed for D3M dataset that has collections
         """
-        suffix = dataset.split('/')[-1]
+        suffix = split(dataset)[-1]
 
         if is_d3m_format(dataset, suffix):
-            dataset = join(dataset, 'dataset_%s/tables/learningData.csv' % suffix)
+            if is_d3m_collection(join(dataset, 'dataset_%s' % suffix, 'datasetDoc.json'), 'text'):
+                dataset = d3mtext_to_dataframe(dataset, text_column)
+            else:
+                dataset = join(dataset, 'dataset_%s' % suffix, 'tables', 'learningData.csv')
 
-        plot_metadata(dataset)
+        vis.plot_metadata(dataset)
 
-    def plot_comparison_pipelines(self, test_dataset=None, source_name=None):
+    def plot_comparison_pipelines(self, test_dataset=None, source_name=None, precomputed_pipelines=None):
         """Plot PipelineProfiler visualization
 
         :param test_dataset: Path to dataset. If None it will use the search scores, otherwise will score the
             pipelines over the passed dataset
         :param source_name: Name of the pipeline source. If None it will use the TA2 id
+        :param precomputed_pipelines: If not None, it loads pipelines previously computed
         """
-        pipelineprofiler_inputs = self.create_pipelineprofiler_inputs(test_dataset, source_name)
-        plot_comparison_pipelines(pipelineprofiler_inputs)
+        if precomputed_pipelines is None:
+            pipelineprofiler_inputs = self.create_pipelineprofiler_inputs(test_dataset, source_name)
+            vis.plot_comparison_pipelines(pipelineprofiler_inputs)
+        else:
+            vis.plot_comparison_pipelines(precomputed_pipelines)
 
-    def plot_text_analysis(self, dataset, text_column, label_column, positive_label=1, negative_label=0):
+    def plot_text_analysis(self, dataset=None, text_column=None, label_column=None, positive_label=1, negative_label=0, precomputed_data=None):
         """Plot a visualization for text datasets
 
         :param dataset: Path to dataset.  It supports D3M dataset, and CSV file
@@ -597,15 +777,13 @@ class AutoML:
         :param label_column: Name of the column that contains the classes
         :param positive_label: Label for the positive class
         :param negative_label: Label for the negative class
+        :param precomputed_data: If not None, it loads words/named entities previously computed
         """
-        suffix = dataset.split('/')[-1]
-
-        if is_d3m_format(dataset, suffix):
-            dataframe = d3mtext_to_dataframe(dataset, text_column)
+        if precomputed_data is not None:
+            vis.plot_text_summary(precomputed_data)
         else:
-            dataframe = pd.read_csv(dataset, index_col=False)
-
-        plot_text_summary(dataframe, text_column, label_column, positive_label, negative_label)
+            precomputed_data = self.create_textanalizer_inputs(dataset, text_column, label_column, positive_label, negative_label)
+            vis.plot_text_summary(precomputed_data)
 
     def plot_text_explanation(self, model_id, instance_text, text_column, label_column, num_features=5, top_labels=1):
         """Plot a LIME visualization for model explanation
@@ -619,8 +797,8 @@ class AutoML:
         """
         train_path = join(self.dataset, 'TRAIN')
         artificial_test_path = join(self.output_folder, 'temp', 'dataset_d3mformat', 'TEST')
-        plot_text_explanation(self, train_path, artificial_test_path, model_id, instance_text, text_column,
-                              label_column, num_features, top_labels)
+        vis.plot_text_explanation(self, train_path, artificial_test_path, model_id, instance_text, text_column,
+                                  label_column, num_features, top_labels)
 
     @staticmethod
     def add_new_ta2(ta2_id, docker_image_url):
@@ -632,4 +810,3 @@ class AutoML:
         """
         TA2_DOCKER_IMAGES[ta2_id] = docker_image_url
         logger.info('%s TA2 added!', ta2_id)
-
